#include "OgreOcclusionCullingSceneManager.h"
#include "OgreMath.h"
#include "OgreIteratorWrappers.h"
#include "OgreRenderSystem.h"
#include "OgreCamera.h"
#include "OgreHardwareOcclusionQuery.h"
//#include "OgreWireBoundingBox.h"
#include "OgreSolidBoundingBox.h"

#include <windows.h>

namespace Ogre {
	
	//-----------------------------------------------------------------------
	OcclusionCullingSceneManager::OcclusionCullingSceneManager(): 
	mFrameId(1), mDistanceQueue(NULL), mVisibilityThreshold(0), mCurrentTestIdx(0),
	mQueryMode(MODE_RENDER), mNumSceneNodes(0)
	{		
	}
	//-----------------------------------------------------------------------
	void OcclusionCullingSceneManager::_findVisibleObjects(Camera* cam, bool onlyShadowCasters)
	{
		// empty because we have to find in _renderVisibleObjects
	}
	//-----------------------------------------------------------------------
	OcclusionCullingSceneManager::~OcclusionCullingSceneManager()
	{		
		deleteQueries();
		//SceneManager::~SceneManager();
	}
	//-----------------------------------------------------------------------
	void OcclusionCullingSceneManager::_renderVisibleObjects(void)
	{
		mDistanceQueue = new PriorityQueue(myless<SceneNode *>(mCameraInProgress));
		
		mDistanceQueue->push(mSceneRoot);

		mCurrentTestIdx = 0;
		//renderZPass();
		//renderCullFrustum();
		//renderCoherentWithQueue();
		renderStopAndWait();
		delete mDistanceQueue;

		/*mDestRenderSystem->_setDepthBufferParams(true, false, CMPF_LESS);
		  SceneManager::_renderVisibleObjects();
		  mDestRenderSystem->_setDepthBufferParams();
          mDistanceQueue.push(mSceneRoot);
		  Preprocess();
		  //ResetQueries();
		*/
		
		mFrameId ++;
	}
	//-----------------------------------------------------------------------
	void OcclusionCullingSceneManager::renderZPass()
	{
		traverseNode(mSceneRoot);
	}
	//-----------------------------------------------------------------------
	void OcclusionCullingSceneManager::renderCoherentWithQueue()
	{
		QueryQueue queryQueue;

		//-- PART 1: process finished occlusion queries
		while(!mDistanceQueue->empty() || !queryQueue.empty())
		{
			while(!queryQueue.empty() && 
				  (queryQueue.front()->getOcclusionQuery()->resultAvailable() || mDistanceQueue->empty()))
			{
				SceneNode *node = queryQueue.front();
				queryQueue.pop();
			
				// wait until result available
				unsigned int visiblePixels;
				node->getOcclusionQuery()->pullOcclusionQuery(&visiblePixels);

				if(visiblePixels > mVisibilityThreshold)
				{
					pullUpVisibility(node);
					traverseNode(node);
				}
			}	

			//-- PART 2: hierarchical traversal
			if(!mDistanceQueue->empty())
			{
				SceneNode *node = mDistanceQueue->top();

				mDistanceQueue->pop();
	
				if(mCameraInProgress->isVisible(node->_getWorldAABB()))
				{
					// identify previously visible nodes
					bool wasVisible = node->isNodeVisible() && (node->lastVisited() == mFrameId - 1);
					
					// identify nodes that we cannot skip queries for
					bool leafOrWasInvisible = !wasVisible || isLeaf(node);

					// reset node's visibility classification 
					node->setVisible(false);

					// update node's visited flag
					node->setLastVisited(mFrameId);
				
					// skip testing previously visible interior nodes
					if(leafOrWasInvisible)
					{
						issueOcclusionQuery(node, wasVisible);
						queryQueue.push(node);
					}
					
					// always traverse a node if it was visible
					if(wasVisible)
						traverseNode(node);
				}
			}
		}
	}
	//-----------------------------------------------------------------------
	void OcclusionCullingSceneManager::renderCullFrustum()
	{
		while(!mDistanceQueue->empty())
		{
			SceneNode *node = mDistanceQueue->top();
			mDistanceQueue->pop();
	
			// interesting for the visualization, so rest and set
			//node->SetVisible(false);
				
			if(mCameraInProgress->isVisible(node->_getWorldAABB()))
			{
				// update node's visited flag => needed for rendering
				// so set it also here
				//node->SetLastVisited(mFrameID);
				//node->SetVisible(true);
				traverseNode(node);
			}
		}
	}	
	//-----------------------------------------------------------------------
	void OcclusionCullingSceneManager::renderStopAndWait()
	{
		while(!mDistanceQueue->empty())
		{
			SceneNode *node = mDistanceQueue->top();
			mDistanceQueue->pop();
		
			// interesting for the visualization
			//node->SetVisible(false);
			node->setLastVisited(mFrameId);

			if(mCameraInProgress->isVisible(node->_getWorldAABB()))
			{
				issueOcclusionQuery(node, false);
				
				unsigned int visiblePixels;
				// wait if result not available
				node->getOcclusionQuery()->pullOcclusionQuery(&visiblePixels);
				
				//char str[100]; sprintf(str, "number: %d, id: %d", (int)visiblePixels, mCurrentTestIdx);
				//MessageBox( NULL, str, "this is my plugin", MB_OK | MB_ICONERROR | MB_TASKMODAL);

				// node visible
				if(visiblePixels > mVisibilityThreshold)
				{
					traverseNode(node);
				}
			}					
		}
	}
	//-----------------------------------------------------------------------
	void OcclusionCullingSceneManager::issueOcclusionQuery(SceneNode *node, bool wasVisible)
	{
		// change state so the bounding box gets not actually rendered on the screen
		setRenderingMode(MODE_QUERY);

		// get next available test id
		HardwareOcclusionQuery *query = mOcclusionQueries[mCurrentTestIdx++];
		node->setOcclusionQuery(query);

		query->beginOcclusionQuery();
		
		SolidBoundingBox boxHalf[2];
		boxHalf[1].mIsFirstHalf = false;

		/*
		static RenderOperation ro;

		// Issue view / projection changes if any
		useRenderableViewProjMode(&box);

		box.getRenderOperation(ro);
		ro.srcRenderable = &box; 
		mDestRenderSystem->_render(ro);*/
		for(int i = 0; i < 2; i++)
		{
			boxHalf[i].setupBoundingBox(node->_getWorldAABB());
			//setPass(boxHalf[i].getTechnique()->getPass(0)); //setRenderingMode(MODE_QUERY);
			SceneManager::renderSingleObject(&boxHalf[i], boxHalf[i].getTechnique()->getPass(0), false);
		}

		query->endOcclusionQuery();
	}

	//-----------------------------------------------------------------------
	void OcclusionCullingSceneManager::setRenderingMode(int mode)
	{	
		// avoid unnecessary state changes
		if(mode != mQueryMode)
		{
			bool enabled = (mode == MODE_RENDER);
			
		//	char str[100]; sprintf(str, "number: %d", mode);
		//	MessageBox( NULL, str, "this is my plugin", MB_OK | MB_ICONERROR | MB_TASKMODAL);
			mDestRenderSystem->_setColourBufferWriteEnabled(enabled, 
				enabled, enabled, enabled);
			mDestRenderSystem->_setDepthBufferWriteEnabled(enabled);
			mDestRenderSystem->setLightingEnabled(enabled);
			mQueryMode = mode;
		}
	}
	//-----------------------------------------------------------------------
	void OcclusionCullingSceneManager::traverseNode(SceneNode *node)
	{
		if(isLeaf(node)) // reached leaf
		{
			render(node);
		}
		else // internal node: add children to priority queue for further processing
		{
			Node::ChildNodeIterator it = node->getChildIterator();
					
			while (it.hasMoreElements())
			{ 
				SceneNode* sceneChild = static_cast<SceneNode*>(it.getNext());
				//mDistanceQueue->addRenderable(sceneChild);
                mDistanceQueue->push(sceneChild);
			}
		 }
	}
	//-----------------------------------------------------------------------
	void OcclusionCullingSceneManager::render(SceneNode *node)
	{
		setRenderingMode(MODE_RENDER);

		//HACK (too slow)
		node->_findVisibleObjects(mCameraInProgress, getRenderQueue(), false, 
								  mDisplayNodes, false);
		SceneManager::_renderVisibleObjects();
		getRenderQueue()->clear();
	}
	//-----------------------------------------------------------------------
	void OcclusionCullingSceneManager::_updateSceneGraph(Camera* cam)
	{
		int numnodes = countSceneNodes(mSceneRoot);
		
		//-- initialise occlusion queries.
		if(numnodes != mNumSceneNodes)
		{
			deleteQueries();
			mNumSceneNodes = numnodes;

			for(unsigned int i=0; i < mNumSceneNodes; i++)
			{
				HardwareOcclusionQuery *hw = mDestRenderSystem->createHardwareOcclusionQuery();
				mOcclusionQueries.push_back(hw);
			}
			//char str[100]; sprintf(str, "number: %d", mNumSceneNodes);
			//MessageBox( NULL, str, "this is my plugin", MB_OK | MB_ICONERROR | MB_TASKMODAL);
		}

		SceneManager::_updateSceneGraph(cam);
	}
	//-----------------------------------------------------------------------
	int OcclusionCullingSceneManager::countSceneNodes(SceneNode *node)
	{
		int num = 1;
		if(isLeaf(node))
		{
			return num;
		}

		Node::ChildNodeIterator it = node->getChildIterator();
		
		while (it.hasMoreElements())
		{ 
			SceneNode* sceneChild = static_cast<SceneNode*>(it.getNext());
			num += countSceneNodes(sceneChild);
        }

		return num;
	}
	//-----------------------------------------------------------------------
	int OcclusionCullingSceneManager::isLeaf(SceneNode *node)
	{
		return (node->numChildren() == 0);
	}
	//-----------------------------------------------------------------------
	void OcclusionCullingSceneManager::pullUpVisibility(SceneNode *node)
	{
		while(node && !node->isNodeVisible())
		{
			node->setNodeVisible(true);
			node = static_cast<SceneNode *>(node->getParent());
		}
	}
	//-----------------------------------------------------------------------
	void OcclusionCullingSceneManager::deleteQueries()
	{
		for(unsigned int i=0; i < mNumSceneNodes; i++)
			delete mOcclusionQueries[i];

		mOcclusionQueries.clear();
	}
}	
